-- vim: ts=2 sw=2 ai et enc=utf8 fenc=utf8

-- Use the table provided for us
local ADDONNAME, RC = ...

-- Make our root addon table accessible via /run etc for easier debugging
RaidChecklist = RC

-- Pull in the localized strings
local L = RaidChecklistStrings
local P = RaidChecklistPetFamilies

if not RaidChecklistSupportedLocale then
  DEFAULT_CHAT_FRAME:AddMessage ("|cffff0000>> Raid Checklist is not supported for your locale ("..GetLocale()..") yet, sorry")
  RaidChecklist = nil
  return
end

RC.colorcodedbackdrop = true


-- The followings strings in the buffs and debuffs arrays are used as keys and should be localized
RC.buffs = {
  "CRIT", "ATTACKPWR", "SPELLPWR6", "SPELLPWR10",
  "PHYSHASTE", "SPELLHASTE",
  "STATS", "ALLDMG", "AGISTR", "STAM", "ARMOR",
  "MANA", "MANAREGEN", "REPLENISHMENT", "BURSTMANA",
  "PUSHBACKRES", "BLOODLUST",
}

RC.debuffs = {
  "SPELLCRIT",
  "PHYSDMGTAKEN", "SPELLDMGTAKEN", "BLEEDDMGTAKEN",
  "ATTACKSPD", "CASTSPD", "DEMOROAR",
  "SUNDER", "MORTALSTRIKE",
}
-- Note: Improved Fairy Fire also has a +SpellHitChance debuff, but seems alone in providing that?

RC.dispels = {
  "OFFMAGIC", "DEFMAGIC",
  "DISEASE", "POISON",
  "CURSE", "ENRAGE"
}

RC.ccs = {
  "CCBEASTS", "CCDEMONS", "CCDRAGONKIN", "CCELEMENTALS", "CCGIANTS", "CCHUMANOIDS", "CCUNDEAD", "CCANY",
  "FEARBEASTS", "FEARUNDEAD", "FEARANY"
}

RC.miscs = {
  "KNOCKBACK", "INTERRUPT", "SNARE", "ROOT", "DISARM"
}

-- Abuse the metatable functionality in order to make sure any abilities referenced by the providers actually exists.
local ability_typo_checker = { __index = function (table, key)
  for i,b in ipairs (table) do
    if b == key then return key end
  end
  DEFAULT_CHAT_FRAME:AddMessage ("|cffff0000>> RaidChecklist: unknown ability name: |r"..(key or "nil"))
end
}
setmetatable (RC.buffs, ability_typo_checker)
setmetatable (RC.debuffs, ability_typo_checker)
setmetatable (RC.dispels, ability_typo_checker)
setmetatable (RC.ccs, ability_typo_checker)
setmetatable (RC.miscs, ability_typo_checker)


-- These aren't localized
local SPEC = " " local NL = "\n "

local availablefrom = {
-- Buffs
  STATS         =
                  L.PALADIN..NL..
                  L.DRUID..NL..
                  L.HUNTER..SPEC..L.BEASTMASTER..L.PET..P.SHALESPIDER,
  AGISTR        =
                  L.WARRIOR..NL..
                  L.SHAMAN..NL..
                  L.DEATHKNIGHT..NL..
                  L.HUNTER..L.PET..P.CAT..NL..
                  L.HUNTER..L.PET..P.SPIRITBEAST,
  STAM          =
                  L.PRIEST..NL..
                  L.WARRIOR..NL..
                  L.WARLOCK..L.PET..P.IMP..NL..
                  L.HUNTER..SPEC..L.BEASTMASTER..L.PET..P.SILITHID,
  ALLDMG        =
                  L.HUNTER..SPEC..L.BEASTMASTER..NL..
                  L.MAGE..SPEC..L.ARCANE..NL..
                  L.PALADIN..SPEC..L.RETRIBUTION,
  CRIT          =
                  L.DRUID..SPEC..L.FERAL..NL..
                  L.WARRIOR..SPEC..L.FURY..NL..
                  L.ROGUE..SPEC..L.SUBTLETY..NL..
                  L.SHAMAN..SPEC..L.ELEMENTAL..NL..
                  L.HUNTER..L.PET..P.WOLF..NL..
                  L.HUNTER..SPEC..L.BEASTMASTER..L.PET..P.DEVILSAUR,
  ATTACKPWR     =
                  L.PALADIN..NL..
                  L.DEATHKNIGHT..SPEC..L.BLOOD..NL..
                  L.HUNTER..SPEC..L.MARKSMAN..NL..
                  L.SHAMAN..SPEC..L.ENHANCEMENT,
  SPELLPWR6     =
                  L.MAGE..NL..
                  L.SHAMAN,
  SPELLPWR10    =
                  L.SHAMAN..SPEC..L.ELEMENTAL..NL..
                  L.WARLOCK..SPEC..L.DEMONOLOGY,
  PHYSHASTE     =
                  L.SHAMAN..NL..
                  L.DEATHKNIGHT..SPEC..L.FROST..NL..
                  L.HUNTER..SPEC..L.SURVIVAL,
  SPELLHASTE    =
                  L.SHAMAN..NL..
                  L.DRUID..SPEC..L.BALANCE..NL..
                  L.PRIEST..SPEC..L.SHADOW,
  BLOODLUST     =
                  L.SHAMAN..NL..
                  L.MAGE..NL..
                  L.HUNTER..SPEC..L.BEASTMASTER..L.PET..P.COREHOUND,
  MANA          =
                  L.MAGE..NL..
                  L.WARLOCK..L.PET..P.FELHUNTER,
  MANAREGEN     =
                  L.PALADIN..NL..
                  L.SHAMAN..NL..
                  L.WARLOCK..L.PET..P.FELHUNTER,
  REPLENISHMENT =
                  L.PRIEST..SPEC..L.SHADOW..NL..
                  L.MAGE..SPEC..L.FROST..NL..
                  L.PALADIN..SPEC..L.RETRIBUTION..NL..
                  L.DRUID..SPEC..L.RESTORATION..NL..
                  L.WARLOCK..SPEC..L.DESTRUCTION,
  BURSTMANA     =
                  L.PRIEST..NL..
                  L.DRUID..SPEC..L.BALANCE..NL..
                  L.DRUID..SPEC..L.RESTORATION..NL..
                  L.SHAMAN..SPEC..L.RESTORATION,
  ARMOR         =
                  L.PALADIN..NL..
                  L.SHAMAN,
  PUSHBACKRES   =
                  L.PALADIN..NL..
                  L.SHAMAN,

-- Debuffs
  SPELLCRIT     =
                  L.MAGE..SPEC..L.FIRE..NL..
                  L.WARLOCK..SPEC..L.DESTRUCTION,
  PHYSDMGTAKEN  =
                  L.WARRIOR..SPEC..L.ARMS..NL..
                  L.ROGUE..SPEC..L.COMBAT..NL..
                  L.DEATHKNIGHT..SPEC..L.FROST..NL..
                  L.HUNTER..L.PET..P.RAVAGER..NL..
                  L.HUNTER..SPEC..L.BEASTMASTER..L.PET..P.WORM,
  SPELLDMGTAKEN =
                  L.WARLOCK..NL..
                  L.ROGUE..SPEC..L.ASSASSINATION..NL..
                  L.DEATHKNIGHT..SPEC..L.UNHOLY..NL..
                  L.DRUID..SPEC..L.BALANCE..NL..
                  L.HUNTER..L.PET..P.DRAGONHAWK..NL..
                  L.HUNTER..L.PET..P.WINDSERPENT,
  BLEEDDMGTAKEN =
                  L.DRUID..SPEC..L.FERAL..NL..
                  L.ROGUE..SPEC..L.SUBTLETY..NL..
                  L.WARRIOR..SPEC..L.ARMS..NL..
                  L.HUNTER..L.PET..P.HYENA..NL..
                  L.HUNTER..L.PET..P.BOAR..NL..
                  L.HUNTER..SPEC..L.BEASTMASTER..L.PET..P.RHINO,
  ATTACKSPD     =
                  L.DEATHKNIGHT..NL..
                  L.SHAMAN..NL..
                  L.WARRIOR..SPEC..L.PROTECTION..NL..
                  L.DRUID..SPEC..L.FERAL..NL..
                  L.PALADIN..SPEC..L.PROTECTION..NL..
                  L.HUNTER..L.PET..P.FOX..NL..
                  L.HUNTER..L.PET..P.TALLSTRIDER,
  CASTSPD       =
                  L.WARLOCK..NL..
                  L.ROGUE..NL..
                  L.DEATHKNIGHT..NL..
                  L.MAGE..SPEC..L.ARCANE..NL..
                  L.HUNTER..L.PET..P.SPOREBAT..NL..
                  L.HUNTER..SPEC..L.BEASTMASTER..L.PET..P.COREHOUND,
  DEMOROAR      =
                  L.WARLOCK..NL..
                  L.WARRIOR..SPEC..L.PROTECTION..NL..
                  L.DRUID..SPEC..L.DRUIDBEAR..NL..
                  L.PALADIN..SPEC..L.PROTECTION..NL..
                  L.DEATHKNIGHT..SPEC..L.BLOOD..NL..
                  L.HUNTER..L.PET..P.CARRIONBIRD..NL..
                  L.HUNTER..L.PET..P.BEAR,
  SUNDER        =
                  L.ROGUE..NL..
                  L.WARRIOR..SPEC..L.PROTECTION..NL..
                  L.DRUID..SPEC..L.DRUIDBEAR..NL..
                  L.HUNTER..L.PET..P.SERPENT..NL..
                  L.HUNTER..L.PET..P.RAPTOR,
  MORTALSTRIKE  =
                  L.ROGUE..NL..
                  L.HUNTER..NL..
                  L.WARRIOR..SPEC..L.ARMS..NL..
                  L.WARRIOR..SPEC..L.FURY..NL..
                  L.PRIEST..SPEC..L.SHADOW..NL..
                  L.MAGE..SPEC..L.FROST..NL..
                  L.WARLOCK..SPEC..L.DEMONOLOGY..L.PET..P.FELGUARD..NL..
                  L.HUNTER..SPEC..L.BEASTMASTER..L.PET..P.DEVILSAUR,
-- Dispel
  OFFMAGIC      =
                  L.PRIEST..NL..
                  L.HUNTER..NL..
                  L.SHAMAN,
  DEFMAGIC      =
                  L.PRIEST..NL..
                  L.PALADIN..SPEC..L.HOLY..NL..
                  L.DRUID..SPEC..L.RESTORATION..NL..
                  L.SHAMAN..SPEC..L.RESTORATION,
  DISEASE       =
                  L.PALADIN..NL..
                  L.PRIEST..SPEC..L.HOLY..NL..
                  L.PRIEST..SPEC..L.DISCIPLINE,
  POISON        =
                  L.DRUID..SPEC..L.BALANCE..NL..
                  L.DRUID..SPEC..L.RESTORATION..NL..
                  L.PALADIN,
  CURSE         =
                  L.SHAMAN..NL..
                  L.MAGE..NL..
                  L.DRUID..SPEC..L.BALANCE..NL..
                  L.DRUID..SPEC..L.RESTORATION,
  ENRAGE        =
                  L.HUNTER..NL..
                  L.ROGUE..NL..
                  L.DRUID..SPEC..L.BALANCE..NL..
                  L.DRUID..SPEC..L.RESTORATION,

-- CC
  CCBEASTS      =
                  L.MAGE..NL..
                  L.SHAMAN..NL..
                  L.ROGUE..NL..
                  L.DRUID..SPEC..L.BALANCE..NL..
                  L.DRUID..SPEC..L.RESTORATION,
  CCDEMONS      =
                  L.WARLOCK..NL..
                  L.ROGUE..NL..
                  L.PALADIN..SPEC..L.RETRIBUTION,
  CCDRAGONKIN   =
                  L.ROGUE..NL..
                  L.DRUID..SPEC..L.BALANCE..NL..
                  L.DRUID..SPEC..L.RESTORATION..NL..
                  L.PALADIN..SPEC..L.RETRIBUTION,
  CCELEMENTALS  =
                  L.WARLOCK..NL..
                  L.SHAMAN,
  CCGIANTS      =
                  L.PALADIN..SPEC..L.RETRIBUTION,
  CCHUMANOIDS   =
                  L.MAGE..NL..
                  L.SHAMAN..NL..
                  L.ROGUE..NL..
                  L.PALADIN..SPEC..L.RETRIBUTION..NL..
                  L.WARLOCK..L.PET..P.SUCCUBUS,
  CCUNDEAD      =
                  L.PRIEST..NL..
                  L.PALADIN..SPEC..L.RETRIBUTION,
  CCANY         =
                  L.HUNTER,
  FEARBEASTS    =
                  L.HUNTER,
  FEARUNDEAD    =
                  L.PALADIN,
  FEARANY       =
                  L.WARLOCK,

-- Misc
  KNOCKBACK     =
                  L.DRUID..SPEC..L.BALANCE..NL..
                  L.WARLOCK..L.PET..P.SUCCUBUS,
  INTERRUPT     =
                  L.DEATHKNIGHT..NL..
                  L.WARRIOR..NL..
                  L.MAGE..NL..
                  L.ROGUE..NL..
                  L.SHAMAN..NL..
                  L.HUNTER..SPEC..L.MARKSMAN..NL..
                  L.WARLOCK..L.PET..P.FELHUNTER,
  SNARE =
                  L.DEATHKNIGHT..NL..
                  L.HUNTER..NL..
                  L.ROGUE..NL..
                  L.SHAMAN..NL..
                  L.WARRIOR..NL..
                  L.MAGE..NL..
                  L.WARLOCK..SPEC..L.AFFLICTION..NL..
                  L.DRUID..SPEC..L.FERAL..NL..
                  L.PRIEST..SPEC..L.SHADOW..NL,
  ROOT =
                  L.DRUID..SPEC..L.BALANCE..NL..
                  L.DRUID..SPEC..L.RESTORATION,
  DISARM =
                  L.WARRIOR..NL..
                  L.ROGUE..NL..
                  L.PRIEST..SPEC..L.SHADOW..NL..
                  L.HUNTER..L.PET..P.BIRDOFPREY..NL..
                  L.HUNTER..L.PET..P.SCORPID,
}


-- No point recomputing these strings all the time, so pre-cache them
local unitstrings = {
  raid = {},
  party = { "player" }, -- GetNumPartyMembers() excludes player!
  player = { "player" }
}
for i = 1,40 do table.insert (unitstrings.raid, "raid"..i) end
for i = 1,4  do table.insert (unitstrings.party, "party"..i) end

-- Find the unit string for the GUID, or nil if not found ("shouldn't happen")
local guid2unit = function (guid)
  local units
  if GetNumRaidMembers () > 0 then
    units = unitstrings.raid
  elseif GetNumPartyMembers () > 0 then
    units = unitstrings.party
  else
    units = unitstrings.player
  end
  for i,unit in ipairs (units) do
    if UnitExists (unit) and UnitGUID (unit) == guid then
      return unit
    end
  end
end


-- The button for anchoring and accessing the checklist
RC.button = CreateFrame ("Button", nil, UIParent)
RC.button:SetNormalTexture ("Interface\\AddOns\\RaidChecklist\\Textures\\button")
RC.button:SetHighlightTexture ("Interface\\AddOns\\RaidChecklist\\Textures\\button")
RC.button:SetPoint ("CENTER", 0, 0)
RC.button:SetWidth (24)
RC.button:SetHeight (24)
RC.button:SetMovable (true)
RC.button:SetUserPlaced (true)
RC.button:SetClampedToScreen (true)
RC.button.OnClick = function (self, button)
  if button == "RightButton" then
    InterfaceOptionsFrame_OpenToCategory (L["Raid Checklist"])
  else
    if RC.list:IsVisible () then
      RC.list:Hide ()
    else
      RC.list:Show ()
    end
  end
end
RC.button:SetScript ("OnClick", RC.button.OnClick)
RC.button:RegisterForClicks ("LeftButtonUp", "RightButtonUp")
RC.button:SetScript ("OnEvent", function (self, event, ...)
  if self[event] then
    self[event] (self, ...)
  end
end)


-- The main list frame
RC.list = CreateFrame ("Frame", "RaidChecklistList", UIParent)
local INSET = 5
RC.list:SetBackdrop ({
  edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
  edgeSize = 16,
  insets= { left = INSET, right = INSET, top = INSET, bottom = INSET }
})
RC.list.texture = RC.list:CreateTexture (nil, "BACKGROUND")
RC.list.texture:SetTexture (0, 0, 0, 0.75)
RC.list.texture:SetPoint ("TOPLEFT", RC.list, "TOPLEFT", INSET, -INSET)
RC.list.texture:SetPoint ("BOTTOMRIGHT", RC.list, "BOTTOMRIGHT", -INSET, INSET)
RC.list:SetPoint ("TOPLEFT", RC.button, "BOTTOMLEFT", 0, 0)
RC.list:Hide ()

-- A frame to hold the buff entries
RC.bufflist = CreateFrame ("Frame", nil, RC.list)
RC.bufflist:SetPoint ("TOPLEFT", RC.list, "TOPLEFT", INSET, -INSET)
RC.bufflist:SetPoint ("BOTTOM", RC.list, "BOTTOM")

-- A frame to hold the debuff entries
RC.debufflist = CreateFrame ("Frame", nil, RC.list)
RC.debufflist:SetPoint ("TOPLEFT", RC.bufflist, "TOPRIGHT")
RC.debufflist:SetPoint ("BOTTOM", RC.list, "BOTTOM")

-- A frame to hold the CC/PvP entries
RC.cclist = CreateFrame ("Frame", nil, RC.list)
RC.cclist:SetPoint ("TOPRIGHT", RC.list, "TOPRIGHT", -INSET, -INSET)
RC.cclist:SetPoint ("LEFT", RC.debufflist, "RIGHT")
RC.cclist:SetPoint ("BOTTOM", RC.list, "BOTTOM")

if (RC.colorcodedbackdrop) then
  local texture = RC.bufflist:CreateTexture (nil, "BACKGROUND")
  texture:SetTexture (0, 0, 1, 0.05)
  texture:SetAllPoints ()
  texture = RC.debufflist:CreateTexture (nil, "BACKGROUND")
  texture:SetTexture (1, 0, 0, 0.05)
  texture:SetAllPoints ()
  texture = RC.cclist:CreateTexture (nil, "BACKGROUND")
  texture:SetTexture (1, 1, 0, 0.05)
  texture:SetAllPoints ()
end

-- The "haven't yet seen talents for" frame
RC.missing = CreateFrame ("Frame", "RaidChecklistMissing", RC.list)
RC.missing:SetBackdrop ({
  edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
  edgeSize = 16,
  insets= { left = INSET, right = INSET, top = INSET, bottom = INSET }
})
RC.missing.texture = RC.missing:CreateTexture (nil, "BACKGROUND")
RC.missing.texture:SetTexture (0, 0, 0, 0.75)
RC.missing.texture:SetPoint ("TOPLEFT", RC.missing, "TOPLEFT", INSET, -INSET)
RC.missing.texture:SetPoint ("BOTTOMRIGHT", RC.missing, "BOTTOMRIGHT", -INSET, INSET)
RC.missing:SetPoint ("TOPLEFT", RC.list, "BOTTOMLEFT")
RC.missing:SetPoint ("RIGHT", RC.list, "RIGHT")
RC.missing.hdr = RC.missing:CreateFontString (nil, "ARTWORK", "GameFontNormal")
RC.missing.hdr:SetPoint ("TOPLEFT", RC.missing, "TOPLEFT", INSET, -INSET)
RC.missing.hdr:SetPoint ("RIGHT", RC.missing, "RIGHT", -INSET, 0)
RC.missing.hdr:SetText (L.MISSINGTALENTS)
RC.missing.txt = RC.missing:CreateFontString (nil, "ARTWORK", "GameFontNormal")
RC.missing.txt:SetPoint ("TOPLEFT", RC.missing.hdr, "BOTTOMLEFT", INSET, 0)
RC.missing.txt:SetPoint ("RIGHT", RC.missing.hdr, "RIGHT")
RC.missing.txt:SetWordWrap (true)
RC.missing.txt:SetTextColor (0.7, 0.7, 0.7, 1)
RC.missing:SetScript ("OnEnter", function ()
  GameTooltip:SetOwner (RC.missing, "ANCHOR_RIGHT", 0, -RC.missing:GetHeight ())
  GameTooltip:SetText (L.INSPECTQUEUE)
  GameTooltip:AddLine (string.gsub (RC.missing_talents_for, ",", "\n"), 0.7, 0.7, 0.7, 1)
  if RC.missing.nowinspecting then
    GameTooltip:AddLine (L.NOWINSPECTING)
    GameTooltip:AddLine (RC.missing.nowinspecting, 0.7, 0.7, 0.7, 1)
  end
  GameTooltip:Show ()
end)
RC.missing:SetScript ("OnLeave", function () GameTooltip:Hide () end)
RC.missing:Hide ()

-- Inspection-in-progress notification
RC.missing.inspecting = CreateFrame ("Frame", nil, RC.missing)
RC.missing.inspecting:SetPoint ("TOPLEFT", RC.missing.hdr, "TOPLEFT")
RC.missing.inspecting:SetSize (12, 12)
RC.missing.inspecting.texture = RC.missing.inspecting:CreateTexture (nil, "ARTWORK")
RC.missing.inspecting.texture:SetTexture ("Interface\\AddOns\\RaidChecklist\\Textures\\light")
RC.missing.inspecting.texture:SetAllPoints ()
RC.missing.inspecting:Hide ()
RC.missing.inspecting:SetScript ("OnEvent", function (self, event, ...)
  if self[event] then
    self[event] (self, ...)
  end
end)
RC.missing.inspecting.INSPECT_READY = function (self)
  --UIFrameFlashStop (self)
  self:Hide ()
  RC.missing.nowinspecting = nil
end
RC.missing.inspecting:RegisterEvent ("INSPECT_READY")
hooksecurefunc("NotifyInspect", function (unit)
  -- UIFrameFlash causes taint apparently, grrr!
  RC.missing.inspecting:Show ()
  -- Flash for 2 minutes, and if we haven't cleared it by then, leave it on
  --UIFrameFlash (RC.missing.inspecting, 0, 1, 120, true, 0, 0)
  RC.missing.nowinspecting = UnitName (unit)
end)



local function merge_providers ()
  local all = {}
  for k,v in pairs (RC.providers) do
    all[k] = {}
    for guid,_ in pairs (v) do
      all[k][guid] = true
    end
  end
  for k,v in pairs (RC.petproviders) do
    for guid,_ in pairs (v) do
      all[k][guid] = true
    end
  end
  return all
end
RC.merge_providers = merge_providers


local function prepare_tooltip (tag)
  GameTooltip:SetText (L[tag], 1, 1, 1, 1)
  GameTooltip:AddLine (L.AVAILABLEFROM)
  GameTooltip:AddLine (" "..availablefrom[tag])
  GameTooltip:AddLine ("\n"..L.PROVIDEDBY, 1, 0.82, 0, 1)
  local guidlist = merge_providers ()[tag]
  if not next (guidlist) then
    GameTooltip:AddLine (" "..L.NOBODY, 0.7, 0.7, 0.7, 1)
  else
    local names
    for guid,_ in pairs (guidlist) do
      if names then names = names.."\n " else names = " " end
      names = names..(RC.provider_names[guid] or ("<missing name for "..guid..">"))
      local unit = guid2unit (guid)
      if unit and not UnitIsConnected (unit) then
        names = names.."|cffff0000 "..L.OFFLINE
      end
    end
    GameTooltip:AddLine (names, 1, 1, 1, 1)
  end
end

local function hide_tooltip ()
  GameTooltip:Hide ()
end


local function make_frame_entry (parent, relto, xoffs, hilight, text, tag)
  local e = CreateFrame ("Frame", nil, parent)
  e:SetPoint ("TOPLEFT", relto, "BOTTOMLEFT")
  local txt = e:CreateFontString (nil, "ARTWORK", "GameFontNormal")
  e:SetWidth (1000) -- Give it enough space to give us a good estimate
  txt:SetPoint ("TOPLEFT", e, "TOPLEFT", xoffs, 0)
  txt:SetPoint ("RIGHT", e, "RIGHT")
  txt:SetJustifyH ("LEFT")
  txt:SetTextColor (0.7, 0.7, 0.7, 1)
  txt:SetText (text)
  e.desired_width = txt:GetStringWidth () * 1.06 + xoffs -- It seems to underestimate :(
  e:SetHeight (txt:GetStringHeight ())
  e:SetPoint ("RIGHT", parent, "RIGHT") -- What we actually want
  e:EnableMouse (true)
  if hilight then
    local tex = e:CreateTexture (nil, "HIGHLIGHT")
    tex:SetAllPoints ()
    tex:SetTexture ("Interface\\Tooltips\\UI-Tooltip-Background") -- FIXME
    tex:SetAlpha (0.3)
    tex:SetVertexColor (1, 1, 0, 0.8)
    e:SetScript ("OnEnter", function ()
      GameTooltip:SetOwner (e, "ANCHOR_RIGHT", 0, -e:GetHeight ())
      prepare_tooltip (e.tag)
      GameTooltip:Show ()
    end)
    e:SetScript ("OnLeave", hide_tooltip)
  end
  e.txt = txt
  e.tag = tag
  return e
end


local function build_lists ()
  RC.buffhdr = make_frame_entry (RC.bufflist, RC.bufflist, 0, false, L.BUFFS)
  RC.buffhdr:SetPoint ("TOPLEFT", RC.bufflist, "TOPLEFT") -- Reposition
  RC.buffhdr.txt:SetTextColor (1, 0.82, 0, 1)

  RC.entries = {}
  local linkto = RC.buffhdr
  local maxwidth = RC.buffhdr.desired_width
  for i,b in ipairs (RC.buffs) do
    local e = make_frame_entry (RC.bufflist, linkto, 8, true, L[b], b)
    maxwidth = max (maxwidth, e.desired_width)
    RC.entries[b] = e
    linkto = e
  end
  RC.bufflist:SetWidth (maxwidth)

  RC.debuffhdr = make_frame_entry (RC.debufflist, RC.buffhdr, 0, false, L.DEBUFFS)
  RC.debuffhdr:SetPoint ("TOPLEFT", RC.buffhdr, "TOPRIGHT") -- Reposition to the right of Buffs list
  RC.debuffhdr.txt:SetTextColor (1, 0.82, 0, 1)
  linkto = RC.debuffhdr
  local maxwidth2 = RC.debuffhdr.desired_width
  for i,b in ipairs (RC.debuffs) do
    local e = make_frame_entry (RC.debufflist, linkto, 8, true, L[b], b)
    maxwidth2 = max (maxwidth2, e.desired_width)
    RC.entries[b] = e
    linkto = e
  end
  linkto = make_frame_entry (RC.debufflist, linkto, 0, false, " ") -- Spacer
  RC.dispelhdr = make_frame_entry (RC.debufflist, linkto, 0, false, L.DISPELS)
  RC.dispelhdr.txt:SetTextColor (1, 0.82, 0, 1)
  linkto = RC.dispelhdr
  for i,b in ipairs (RC.dispels) do
    local e = make_frame_entry (RC.debufflist, linkto, 8, true, L[b], b)
    maxwidth2 = max (maxwidth2, e.desired_width)
    RC.entries[b] = e
    linkto = e
  end
  RC.debufflist:SetWidth (maxwidth2)

  RC.cchdr = make_frame_entry (RC.cclist, RC.debuffhdr, 0, false, L.CC)
  RC.cchdr:SetPoint ("TOPLEFT", RC.debuffhdr, "TOPRIGHT") -- Reposition to the right of Debuffs list
  RC.cchdr.txt:SetTextColor (1, 0.82, 0, 1)
  linkto = RC.cchdr
  local maxwidth3 = RC.cchdr.desired_width
  local totalheight = RC.cchdr:GetHeight ()
  for i,b in ipairs (RC.ccs) do
    local e = make_frame_entry (RC.cclist, linkto, 8, true, L[b], b)
    maxwidth3 = max (maxwidth3, e.desired_width)
    totalheight = totalheight + e:GetHeight ()
    RC.entries[b] = e
    linkto = e
  end
  linkto = make_frame_entry (RC.cclist, linkto, 0, false, " ") -- Spacer
  totalheight = totalheight + linkto:GetHeight ()
  RC.mischdr = make_frame_entry (RC.cclist, linkto, 0, false, L.MISC)
  RC.mischdr.txt:SetTextColor (1, 0.82, 0, 1)
  linkto = RC.mischdr
  totalheight = totalheight + linkto:GetHeight ()
  for i,b in ipairs (RC.miscs) do
    local e = make_frame_entry (RC.cclist, linkto, 8, true, L[b], b)
    maxwidth3 = max (maxwidth3, e.desired_width)
    totalheight = totalheight + e:GetHeight ()
    RC.entries[b] = e
    linkto = e
  end
  RC.cclist:SetWidth (maxwidth3)

  RC.list:SetWidth (maxwidth + maxwidth2 +maxwidth3 + 2 * INSET)
  RC.list:SetHeight (totalheight + 3 * INSET) -- Looks better with a bit of extra space
end

build_lists ()


-- Levels: 0 = missing, 1 = offline, 2 = filtered, 3 = provided
local providerlevel4guid = function (guid)
  local db = RaidChecklistDB
  local level = 0
  local unit = guid2unit (guid)
  if unit then
    level = 3
    if not UnitIsConnected (unit) then
      level = 1
    end
    local israid, _, idx = string.find (unit, "raid(%d+)")
    if israid then
      local group = "group"..select(3, GetRaidRosterInfo (idx))
      if not db.profile[group] then
        level = 2
      end
      group = nil
    end
  end
  return level
end

local providerlevel = function (guidlist)
  local highestlevel = 0
  for guid,_ in pairs (guidlist) do
    local level = providerlevel4guid (guid)
    if level > highestlevel then highestlevel = level end
  end
  return highestlevel
end

-- Updates the color on each of the buff entries in the list
RC.OnRefresh = function (self)
  local all = self.merge_providers ()
  local colors = {
    [0] = { r = 0.7, g = 0.7, b = 0.7 },
    [1] = { r = 0.7, g = 0.0, b = 0.0 },
    [2] = { r = 0.0, g = 0.4, b = 1.0 },
    [3] = { r = 0.0, g = 1.0, b = 0.0 }
  }
  for key,guidlist in pairs (all) do
    local plevel = providerlevel (guidlist)
    local c = colors[plevel]
    self.entries[key].txt:SetTextColor (c.r, c.g, c.b, 1)
  end
  all = nil

  -- Show players we haven't seen talents for yet, if there are any
  if self.missing_talents_for then
    self.missing.txt:SetText (string.gsub (self.missing_talents_for, ",", ", "))
    self.missing:SetHeight (self.missing.hdr:GetHeight () * 2 + self.missing.txt:GetHeight ())
    self.missing:Show ()
  else
    self.missing:Hide ()
  end
end


RC.OnConfigChange = function (self)
  local db = RaidChecklistDB
  if db.profile.show_anchor then
    self.button:Show ()
  else
    self.button:Hide ()
  end

  if db.profile.anchor_locked then
    self.button:SetScript ("OnMouseDown", nil)
    self.button:SetScript ("OnMouseUp", nil)
  else
    self.button:SetScript ("OnMouseDown", function (f) f:StartMoving () end)
    self.button:SetScript ("OnMouseUp", function (f)
      f:StopMovingOrSizing ()
      db.profile.anchorpoint, _, db.profile.relpoint, db.profile.x, db.profile.y = f:GetPoint (1)
    end)
  end

  self.list:ClearAllPoints ()
  self.missing:ClearAllPoints ()
  self.missing:SetPoint ("RIGHT", self.list, "RIGHT")
  local dir = db.profile.open_direction
  if dir == "DOWNRIGHT" then
    self.list:SetPoint ("TOPLEFT", self.button, "BOTTOMLEFT", 0, 0)
    self.missing:SetPoint ("TOPLEFT", self.list, "BOTTOMLEFT")
  elseif dir == "DOWNLEFT" then
    self.list:SetPoint ("TOPRIGHT", self.button, "BOTTOMRIGHT", 0, 0)
    self.missing:SetPoint ("TOPLEFT", self.list, "BOTTOMLEFT")
  elseif dir == "UPRIGHT" then
    self.list:SetPoint ("BOTTOMLEFT", self.button, "TOPLEFT", 0, 0)
    self.missing:SetPoint ("BOTTOMLEFT", self.list, "TOPLEFT")
  else -- UPLEFT
    self.list:SetPoint ("BOTTOMRIGHT", self.button, "TOPRIGHT", 0, 0)
    self.missing:SetPoint ("BOTTOMLEFT", self.list, "TOPLEFT")
  end

  self.button:SetPoint (db.profile.anchorpoint, UIParent, db.profile.relpoint, db.profile.x, db.profile.y)

  local fontfile, fontsize = self.entries.CRIT.txt:GetFont ()
  if (fontsize < 11 and db.profile.font_size == "normal") or (fontsize > 11 and db.profile.font_size == "small") then
    local grow = db.profile.font_size == "normal"
    if db.profile.font_size == "normal" then
      fontsize = 12
    elseif db.profile.font_size == "small" then
      fontsize = 10
    end

    for _,e in pairs (RC.entries) do
      e.txt:SetFont (fontfile, fontsize)
      e:SetHeight (fontsize)
    end
    local hdrs = { RC.buffhdr, RC.debuffhdr, RC.dispelhdr, RC.cchdr, RC.mischdr }
    for _,e in pairs (hdrs) do
      e.txt:SetFont (fontfile, fontsize)
      e:SetHeight (fontsize)
    end

    local frames = { RC.bufflist, RC.debufflist, RC.cclist, RC.list }
    if grow then
      for _, frame in ipairs (frames) do
        frame:SetWidth (frame:GetWidth () * 1.2)
      end
      RC.list:SetHeight (RC.list:GetHeight () * 1.2)
    else
      for _, frame in ipairs (frames) do
        frame:SetWidth (frame:GetWidth () / 1.2)
      end
      RC.list:SetHeight (RC.list:GetHeight () / 1.2)
    end
  end

  self:OnRefresh () -- In case the filtered groups changed
end

RC.button.ADDON_LOADED = function (self, name)
  if name == ADDONNAME then
    if not RaidChecklistDB then RaidChecklistDB = {} end
    if not RaidChecklistDB.profile then RaidChecklistDB.profile = {} end
    setmetatable (RaidChecklistDB.profile, { __index = RC.dbdefaults.profile })
    RC:OnConfigChange ()
    self:UnregisterEvent ("ADDON_LOADED")
  end
end
RC.button:RegisterEvent ("ADDON_LOADED")


